/*!
 * ve-charts v1.0.0-alpha.10
 * (c) 2021 xiaoluoboding
 * @license MIT
 */
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var vue = require('vue');
var core = require('echarts/core');
var lodashEs = require('lodash-es');
var resizeObserver = require('@juggle/resize-observer');
var charts = require('echarts/charts');
var components = require('echarts/components');
var renderers = require('echarts/renderers');

/* eslint-disable @typescript-eslint/no-explicit-any */
const toKebabCase = (str) => str
    .replace(/([A-Z])([A-Z])/g, '$1-$2')
    .replace(/([a-z])([A-Z])/g, '$1-$2')
    .replace(/[\s_]+/g, '-')
    .toLowerCase();
const toCamelCase = (arr) => arr
    .map((str, idx) => idx > 0 ? str.charAt(0).toLocaleUpperCase() + str.slice(1) : str)
    .join('');
// Copied from
// https://github.com/vuejs/vue-next/blob/5a7a1b8293822219283d6e267496bec02234b0bc/packages/shared/src/index.ts#L40-L41
const isOn = (key) => /^on[^a-z]/.test(key);
const omitOn = (attrs) => {
    const result = {};
    for (const key in attrs) {
        if (!isOn(key)) {
            result[key] = attrs[key];
        }
    }
    return result;
};
const withInstall = (comp) => {
    const c = comp;
    c.install = (app) => {
        app.component(c.name, c);
    };
    return c;
};
const unrefElement = (refEl) => {
    return vue.unref(refEl);
};
// Returns true if the given value is a number, false otherwise.
const validateNumber = (n) => !isNaN(parseFloat(n)) && isFinite(n) && Number(n) === n;
/**
 * Transform dataset
 * @param data
 * @param settings
 * @param extra
 */
const getDataset = (data, settings, options) => {
    const cloneData = lodashEs.cloneDeep(data);
    const dimName = cloneData?.dimensions?.name;
    const dimData = cloneData?.dimensions?.data;
    const stack = settings?.stack || null;
    const percentage = settings?.percentage || false;
    // when data is not empty and data.dimensions.data is undefiend
    if (!options?.isEmptyData && dimData === undefined) {
        // Vue.util.warn(`data.dimensions.data is required. Please check on you data`, this)
        return;
    }
    /**
     * echarts render problem with dataset
     * problem1: when dimension name a number-like string(e.g., `"1"`), start with number and <= measures.length
     * https://github.com/apache/incubator-echarts/blob/4e4cf884fc8a96b18bd1de537b590042e49df684/src/data/List.js#L339
     * solution: append a ' ' to the dimension name
     * problem2: [piechart] when measure.data start with number
     * solution: use special String instead
     * PS. in version v0.4.6 all seems fixed
     */
    const dimKey = `${dimName}`;
    const headMeasure = dimData.length > 0 && dimData[0];
    const dimValue = validateNumber(headMeasure) && options?.chartType === 'pie'
        ? dimData.map((v, i) => (i === 0 ? `${v}` : v))
        : dimData;
    const dimensions = {
        [dimKey]: dimValue
    };
    // let [measures, zipSumed] = [{}, []]
    const measures = {};
    let zipSumed = [];
    if (stack && percentage && cloneData.measures.length > 0) {
        const dyadicArray = cloneData.measures.map(col => col.data);
        // 横表转竖表 用于计算百分比堆叠图
        const zipped = lodashEs.zip(...dyadicArray);
        zipSumed = zipped.map((v) => {
            const arr = v.map((v) => {
                return validateNumber(v) ? v : parseFloat(v);
            });
            return lodashEs.sum(arr);
        });
    }
    cloneData.measures.map(row => {
        const isNumber = validateNumber(row.name);
        const rowName = isNumber ? `${row.name} ` : row.name;
        Object.assign(measures, {
            [rowName]: stack && percentage
                ? row.data.map((v, i) => lodashEs.round(v / zipSumed[i], 4))
                : row.data
        });
    });
    let dims = [];
    const firstDim = dimName === undefined ? 'dimension' : dimName;
    dims.push(firstDim);
    dims = [...dims, ...cloneData.measures.map(v => v.name)];
    const source = Object.assign({}, dimensions, measures);
    const dataset = {
        dimensions: dims,
        source
    };
    // console.log(JSON.stringify(dataset))
    return dataset;
};

/* eslint-disable @typescript-eslint/no-explicit-any */
const useEchartsEvents = (attrs) => {
    const echartsAttrs = vue.computed(() => omitOn(attrs));
    const echartsEvents = {};
    Object.keys(attrs)
        .filter(key => isOn(key) && typeof attrs[key] === 'function')
        .forEach(key => {
        const [, ...rest] = toKebabCase(key).split('-');
        const eventKey = rest.length === 1
            ? rest.join('')
            : toCamelCase(rest);
        echartsEvents[eventKey] = attrs[key];
    });
    return {
        echartsAttrs,
        echartsEvents
    };
};

/**
 * It immediately detects when an element resizes and provides accurate sizing information back to the handler.
 * @param target
 * @param callback
 * @param options
 */
const useResizeObserver = (target, callback, options = {}) => {
    let ro;
    const isSupported = window && 'ResizeObserver' in window;
    const stopWatch = vue.watch(() => unrefElement(target), (el) => {
        if (isSupported && window && el) {
            ro = new resizeObserver.ResizeObserver(callback);
            ro.observe(el, options);
        }
    }, { immediate: true });
    const stop = () => {
        ro && ro.disconnect();
        ro = undefined;
        stopWatch();
    };
    vue.onUnmounted(stop);
    return {
        stop
    };
};

/* eslint-disable @typescript-eslint/no-explicit-any */
var Chart$9 = vue.defineComponent({
    name: 'VeChart',
    inheritAttrs: false,
    props: {
        option: {
            type: Object,
            default: () => ({})
        },
        initOptions: {
            type: Object,
            default: () => ({
                renderer: 'canvas'
            })
        },
        theme: {
            type: [String, Object],
            default: 'default'
        },
        setOptionOpts: {
            type: Object,
            default: () => ({})
        },
        height: { type: Number, default: 400 },
        needUpdate: { type: Boolean, default: false }
    },
    setup(props, { attrs }) {
        const option = vue.toRef(props, 'option');
        const initOptions = vue.toRef(props, 'initOptions');
        const theme = vue.toRef(props, 'theme');
        const setOptionOpts = vue.toRef(props, 'setOptionOpts');
        // const height = toRef(props, 'height')
        const needUpdate = vue.toRef(props, 'needUpdate');
        const echartsRef = vue.ref();
        const echartsInstance = vue.shallowRef();
        const canvasRect = vue.ref({});
        const { echartsEvents } = useEchartsEvents(attrs);
        // console.log(option.value)
        const resize = () => {
            if (echartsInstance.value) {
                echartsInstance.value.resize();
            }
        };
        const dispose = () => {
            if (echartsInstance.value) {
                echartsInstance.value.dispose();
                echartsInstance.value = undefined;
            }
        };
        const delegateEvents = (instance) => {
            Object.keys(echartsEvents).forEach(key => {
                const handler = echartsEvents[key];
                if (!handler)
                    return;
                if (key.indexOf('zr:') === 0) {
                    instance.getZr().on(key.slice(3).toLowerCase(), handler);
                }
                else {
                    instance.on(key, handler);
                }
            });
        };
        const init = (opt) => {
            if (!echartsRef.value)
                return;
            echartsInstance.value = core.init(echartsRef.value, theme.value, initOptions.value);
            opt && echartsInstance.value.setOption(opt);
            delegateEvents(echartsInstance.value);
            vue.nextTick(resize);
        };
        const setOption = (option, opts) => {
            if (!echartsInstance.value) {
                init(option);
            }
            else {
                echartsInstance.value.setOption(option, opts);
            }
        };
        const isEmptyData = (opt) => lodashEs.isEmpty(opt?.series) && lodashEs.isEmpty(opt?.dataset);
        useResizeObserver(echartsRef, ([entry]) => {
            canvasRect.value = entry.contentRect;
            resize();
        });
        vue.watch(needUpdate, (needUpdate) => {
            // console.log(needUpdate)
            // console.log(JSON.stringify(option.value))
            if (needUpdate && option.value && setOptionOpts.value) {
                setOption(option.value, setOptionOpts.value);
            }
            else {
                setOption(option.value);
            }
        }, { deep: true });
        vue.watch([theme, initOptions], () => {
            dispose();
            init(option.value);
        }, { deep: true });
        vue.onMounted(() => {
            // console.group('--- new chart ---')
            // console.log(JSON.stringify(option.value))
            // console.log(isEmptyData(option.value) ? '空数据' : '有数据')
            if (!isEmptyData(option.value)) {
                init(option.value);
            }
            // console.groupEnd()
        });
        vue.onUnmounted(dispose);
        return {
            echartsRef
        };
    },
    render(ctx) {
        const scopeId = toKebabCase((ctx.$attrs.chartType || 've-chart'));
        vue.pushScopeId(scopeId);
        const echartsStyle = {
            width: 'auto',
            height: `${ctx.$props.height}px`
        };
        return vue.h('div', {
            ref: 'echartsRef',
            style: echartsStyle
        });
    }
});

const VeChart = withInstall(Chart$9);

core.use([
    components.GridComponent,
    renderers.CanvasRenderer
]);

/* eslint-disable @typescript-eslint/no-explicit-any */
// setup can not be mixins or extends
var HocChart = vue.defineComponent({
    props: {
        /**
         * echarts options
         * ve-charts delegate all echarts options as component props
         * https://echarts.apache.org/zh/option.html
         */
        title: Object,
        legend: Object,
        grid: Object,
        xAxis: [Object, Array],
        yAxis: [Object, Array],
        polar: Object,
        radiusAxis: Object,
        angleAxis: Object,
        radar: [Object, Array],
        dataZoom: [Object, Array],
        visualMap: [Object, Array],
        tooltip: Object,
        axisPointer: Object,
        toolbox: Object,
        brush: Object,
        geo: Object,
        parallel: Object,
        parallelAxis: Array,
        singleAxis: Array,
        timeline: Object,
        graphic: Object,
        calendar: Object,
        dataset: Object,
        aria: Object,
        series: [Object, Array],
        darkMode: { type: [Boolean, String], default: 'auto' },
        color: Array,
        backgroundColor: [Object, String],
        textStyle: Object,
        animation: { type: Boolean, default: true },
        animationThreshold: Number,
        animationDuration: [Number, Function],
        animationEasing: [String, Function],
        animationDelay: [Number, Function],
        animationDurationUpdate: [Number, Function],
        animationEasingUpdate: [String, Function],
        animationDelayUpdate: [String, Function],
        stateAnimation: Object,
        blendMode: String,
        hoverLayerThreshold: Number,
        useUTC: { type: Boolean, default: false },
        options: Object,
        media: Array,
        // ve-charts basic props
        data: [Object, Array],
        variant: { type: String, default: '' },
        loading: { type: Boolean, default: false },
        emptyText: { type: String, default: 'Empty Data' },
        // ve-charts common props
        option: {
            type: [Object, Array],
            default: () => ({})
        }
    },
    data: () => ({
        initOptions: {
            renderer: 'canvas'
        },
        needUpdate: false,
        setOptionOpts: {},
        mergedOption: {}
    }),
    computed: {
        chartOpts() {
            return {
                option: this.mergedOption,
                initOptions: this.initOptions,
                needUpdate: this.needUpdate,
                setOptionOpts: this.setOptionOpts,
                chartType: this.$options.name
            };
        },
        isEmptyData() {
            const opt = this.mergedOption;
            return lodashEs.isEmpty(opt?.series) && lodashEs.isEmpty(opt?.dataset);
        }
    },
    created() {
        this.mergedOption = this.option;
        const propsKeys = Object.keys(this.$props);
        if (this.$props && propsKeys.length > 0) {
            propsKeys
                // .forEach((prop: string) => (this.$props[prop] === undefined) && delete this.$props[prop])
                .filter((prop) => !/(data|settings|height|loading|emptyText|variant)$/.test(prop))
                .forEach((prop) => {
                this.$watch(prop, (val) => {
                    if (!lodashEs.isEmpty(val)) {
                        // merge echarts default option
                        const option = prop === 'option'
                            ? this.option
                            : { [prop]: val };
                        this.mergePropsToOption(option);
                    }
                    this.forcedUpdate();
                }, {
                    // TODO only object types need deep watch
                    deep: true,
                    immediate: true
                });
            });
        }
    },
    beforeUnmount() {
        this.mergedOption = {};
    },
    methods: {
        mergePropsToOption(props) {
            const option = lodashEs.cloneDeep(this.mergedOption);
            this.mergedOption = { ...option, ...props };
        },
        forcedUpdate() {
            // manual update echarts options
            this.needUpdate = true;
            this.$nextTick(() => { this.needUpdate = false; });
        }
    },
    render() {
        return vue.h(VeChart, this.chartOpts);
    }
});

core.use([charts.BarChart]);
var Chart$8 = vue.defineComponent({
    name: 'BarChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const BASE_OPTION = {
    grid: {
        right: 10,
        bottom: 10,
        left: 10,
        containLabel: true
    },
    series: []
};

class Bar {
    constructor(props) {
        this.$props = props;
        this.data = this.$props.data;
        this.settings = this.$props.settings;
        // state
        this.isColumn = this.settings?.variant === 'column';
    }
    // build grid
    getBarGrid(isColumn) {
        const columnGrid = {
            right: 30,
            bottom: 10,
            left: 30,
            containLabel: true
        };
        const grid = isColumn ? columnGrid : BASE_OPTION.grid;
        return { ...grid, ...this.$props?.grid };
    }
    getBarDimAxis() {
        return {
            type: 'category'
        };
    }
    getBarMeaAxis() {
        return {
            type: 'value'
        };
    }
    getSeries() {
        let series = [];
        series = this.data.measures.map(({ name }, idx) => {
            const seriesItem = this.$props?.series?.[idx] || {};
            return {
                type: 'bar',
                name,
                ...seriesItem
            };
        });
        return series;
    }
    chartHandler() {
        const xAxis = this.isColumn ? this.getBarMeaAxis() : this.getBarDimAxis();
        const yAxis = this.isColumn ? this.getBarDimAxis() : this.getBarMeaAxis();
        // build echarts options
        const option = {
            grid: this.getBarGrid(this.isColumn),
            xAxis,
            yAxis,
            dataset: getDataset(this.data),
            series: this.getSeries(),
            ...this.settings
        };
        return option;
    }
}

/* eslint-disable @typescript-eslint/no-explicit-any */
core.use([charts.BarChart, components.DatasetComponent]);
var Variants$2 = vue.defineComponent({
    name: 'VeBarChart',
    extends: HocChart,
    props: {
        settings: [Object, Array]
    },
    mounted() {
        const baseOption = new Bar(this.$props).chartHandler();
        console.log(JSON.stringify(baseOption));
        this.mergedOption = { ...this.mergedOption, ...baseOption };
    }
});

const BarChart = withInstall(Chart$8);
const VeBarChart = withInstall(Variants$2);

core.use([charts.LineChart]);
var Chart$7 = vue.defineComponent({
    name: 'LineChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

class Line {
    constructor(props) {
        this.$props = props;
        this.data = this.$props.data;
        this.settings = this.$props.settings;
        // state
        this.chartType = 'line';
    }
    // build grid
    getLineGrid() {
        const grid = BASE_OPTION.grid;
        return { ...grid, ...this.$props?.grid };
    }
    getLineDimAxis() {
        return {
            type: 'category'
        };
    }
    getLineMeaAxis() {
        return {
            type: 'value'
        };
    }
    getSeries() {
        let series = [];
        series = this.data.measures.map(({ name }, idx) => {
            const seriesItem = this.$props?.series?.[idx] || {};
            return {
                type: this.chartType,
                name,
                step: this.settings?.step || false,
                smooth: this.settings?.smooth || false,
                ...seriesItem
            };
        });
        return series;
    }
    chartHandler() {
        const xAxis = this.getLineDimAxis();
        const yAxis = this.getLineMeaAxis();
        // build echarts options
        const option = {
            grid: this.getLineGrid(),
            xAxis,
            yAxis,
            dataset: getDataset(this.data),
            series: this.getSeries(),
            ...this.settings
        };
        return option;
    }
}

/* eslint-disable @typescript-eslint/no-explicit-any */
core.use([charts.LineChart, components.DatasetComponent]);
var Variants$1 = vue.defineComponent({
    name: 'VeLineChart',
    extends: HocChart,
    props: {
        settings: [Object, Array]
    },
    mounted() {
        const baseOption = new Line(this.$props).chartHandler();
        console.log(JSON.stringify(baseOption));
        this.mergedOption = { ...this.mergedOption, ...baseOption };
    }
});

const LineChart = withInstall(Chart$7);
const VeLineChart = withInstall(Variants$1);

core.use([charts.PieChart]);
var Chart$6 = vue.defineComponent({
    name: 'PieChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

class Pie {
    constructor(props) {
        this.$props = props;
        this.data = this.$props.data;
        this.settings = this.$props.settings;
        // state
        this.isDonut = this.settings?.variant === 'donut';
    }
    getPieDataset() {
        const dataset = [];
        const data = this.data;
        const settings = this.settings;
        if (!Array.isArray(data)) {
            dataset.push(getDataset(data, settings, { chartType: 'pie' }));
        }
        else if (Array.isArray(data) && data.length > 1) {
            for (const element of data) {
                dataset.push(getDataset(element, settings, { chartType: 'pie' }));
            }
        }
        return dataset;
    }
    getPieTooltip() {
        return {
            trigger: 'item'
        };
    }
    getPieSeries() {
        let series = [];
        series = this.data.measures.map(({ name }, idx) => {
            const seriesItem = this.$props?.series?.[idx] || {};
            return {
                type: 'pie',
                name,
                radius: this.isDonut ? ['35%', '60%'] : [0, '60%'],
                ...seriesItem
            };
        });
        return series;
    }
    chartHandler() {
        // build echarts options
        const option = {
            tooltip: this.getPieTooltip(),
            dataset: this.getPieDataset(),
            series: this.getPieSeries(),
            ...this.settings
        };
        // console.log(option)
        return option;
    }
}

/* eslint-disable @typescript-eslint/no-explicit-any */
core.use([charts.PieChart, components.DatasetComponent]);
var Variants = vue.defineComponent({
    name: 'VePieChart',
    extends: HocChart,
    props: {
        settings: [Object, Array]
    },
    mounted() {
        const baseOption = new Pie(this.$props).chartHandler();
        console.log(JSON.stringify(baseOption));
        this.mergedOption = { ...this.mergedOption, ...baseOption };
    }
});

const PieChart = withInstall(Chart$6);
const VePieChart = withInstall(Variants);

core.use([charts.FunnelChart]);
var Chart$5 = vue.defineComponent({
    name: 'FunnelChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const FunnelChart = withInstall(Chart$5);

core.use([charts.GaugeChart]);
var Chart$4 = vue.defineComponent({
    name: 'GaugeChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const GaugeChart = withInstall(Chart$4);

core.use([charts.RadarChart]);
var Chart$3 = vue.defineComponent({
    name: 'RadarChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const RadarChart = withInstall(Chart$3);

core.use([charts.ScatterChart]);
var Chart$2 = vue.defineComponent({
    name: 'ScatterChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const ScatterChart = withInstall(Chart$2);

core.use([charts.TreeChart]);
var Chart$1 = vue.defineComponent({
    name: 'TreeChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const TreeChart = withInstall(Chart$1);

core.use([charts.TreemapChart]);
var Chart = vue.defineComponent({
    name: 'TreemapChart',
    extends: HocChart,
    setup(props) {
        console.log(props);
    }
});

const TreemapChart = withInstall(Chart);

const componentList = [
    VeChart,
    BarChart,
    VeBarChart,
    LineChart,
    VeLineChart,
    PieChart,
    VePieChart,
    FunnelChart,
    GaugeChart,
    RadarChart,
    ScatterChart,
    TreeChart,
    TreemapChart
];
const install = (app) => {
    componentList.forEach(component => {
        app.component(component.name, component);
    });
};

exports.BarChart = BarChart;
exports.FunnelChart = FunnelChart;
exports.GaugeChart = GaugeChart;
exports.LineChart = LineChart;
exports.PieChart = PieChart;
exports.RadarChart = RadarChart;
exports.ScatterChart = ScatterChart;
exports.TreeChart = TreeChart;
exports.TreemapChart = TreemapChart;
exports.VeBarChart = VeBarChart;
exports.VeChart = VeChart;
exports.VeLineChart = VeLineChart;
exports.VePieChart = VePieChart;
exports.default = install;
